@@ -297,9 +297,10 @@ macro_rules! impl_base {
             fn reset(&mut self) -> Result<(), EspError> {
                 #[cfg(not(feature = "riscv-ulp-hal"))]
                 let res = {
-                    esp!(unsafe { gpio_reset_pin(self.pin()) })?;
                     #[cfg(feature = "alloc")]
                     self.disable_interrupt()?;
+
+                    esp!(unsafe { gpio_reset_pin(self.pin()) })?;
                     Ok(())
                 };
 
@@ -435,7 +436,7 @@ macro_rules! impl_base {
 }
 
 macro_rules! impl_pull {
-    ($pxi:ident: $mode:ident) => {
+    ($pxi:ident: $mode:ty) => {
         impl Pull for $pxi<$mode> {
             type Error = EspError;
 
@@ -528,25 +529,17 @@ macro_rules! impl_input_base {
                 Ok($pxi { _mode: PhantomData })
             }
 
-            /// Degrades a concrete pin (e.g. [`Gpio1`]) to a generic pin
-            /// struct that can also be used with periphals.
-            pub fn degrade(self) -> GpioPin<MODE> {
-                unsafe { GpioPin::new($pin) }
-            }
-        }
-
-        #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
-        impl $pxi<Input> {
             /// # Safety
             ///
             /// The callback passed to this method is executed in the context of an
             /// interrupt handler. So you should take care of what is done in it.
+            #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
             pub unsafe fn into_subscribed(
                 mut self,
-                callback: impl FnMut() + 'static,
+                callback: impl FnMut() + Send + 'static,
                 interrupt_type: InterruptType,
             ) -> Result<$pxi<SubscribedInput>, EspError> {
-                self.enable_interrupt()?;
+                self.set_input()?;
 
                 self.set_interrupt_type(interrupt_type)?;
 
@@ -554,8 +547,16 @@ macro_rules! impl_input_base {
 
                 register_irq_handler(self.pin() as usize, callback);
 
+                self.enable_interrupt()?;
+
                 Ok($pxi { _mode: PhantomData })
             }
+
+            /// Degrades a concrete pin (e.g. [`Gpio1`]) to a generic pin
+            /// struct that can also be used with periphals.
+            pub fn degrade(self) -> GpioPin<MODE> {
+                unsafe { GpioPin::new($pin) }
+            }
         }
 
         impl<MODE> Pin for $pxi<MODE>
@@ -592,6 +593,10 @@ macro_rules! impl_input_output {
         impl_input_base!($pxi: $pin);
         impl_pull!($pxi: Input);
         impl_pull!($pxi: InputOutput);
+
+        #[cfg(all(not(feature = "riscv-ulp-hal"), feature = "alloc"))]
+        impl_pull!($pxi: SubscribedInput);
+
         impl_hal_input_pin!($pxi: InputOutput);
 
         impl<MODE> OutputPin for $pxi<MODE> where MODE: Send {}
@@ -645,6 +650,25 @@ macro_rules! impl_rtc {
     ($pxi:ident: $pin:expr, NORTC: $rtc:expr) => {};
 }
 
+macro_rules! impl_adc_pull {
+    ($pxi:ident, ADC1) => {
+        impl_adc_pull!($pxi, adc::ADC1);
+    };
+
+    ($pxi:ident, ADC2) => {
+        impl_adc_pull!($pxi, adc::ADC2);
+    };
+
+    ($pxi:ident, NOADC) => {};
+
+    ($pxi:ident, $adc:ty) => {
+        impl_pull!($pxi: adc::Atten0dB<$adc>);
+        impl_pull!($pxi: adc::Atten2p5dB<$adc>);
+        impl_pull!($pxi: adc::Atten6dB<$adc>);
+        impl_pull!($pxi: adc::Atten11dB<$adc>);
+    };
+}
+
 macro_rules! impl_adc {
     ($pxi:ident: $pin:expr, ADC1: $adc:expr) => {
         #[cfg(not(feature = "riscv-ulp-hal"))]
@@ -788,7 +812,7 @@ macro_rules! impl_adc {
             type ID = u8;
 
             fn channel() -> Self::ID {
-                $adc as u8
+                $adc
             }
         }
 
@@ -799,7 +823,7 @@ macro_rules! impl_adc {
             type ID = u8;
 
             fn channel(&self) -> Self::ID {
-                adc_unit_t_ADC_UNIT_2 as u8
+                $adc
             }
         }
     };
@@ -938,6 +962,7 @@ macro_rules! pin {
         impl_input_output!($pxi: $pin);
         impl_rtc!($pxi: $pin, $rtc: $rtcno);
         impl_adc!($pxi: $pin, $adc: $adcno);
+        impl_adc_pull!($pxi, $adc);
         impl_dac!($pxi: $pin, $dac: $dacno);
         impl_touch!($pxi: $pin, $touch: $touchno);
     };
