@@ -1,3 +1,5 @@
+use core::marker::PhantomData;
+use embedded_hal::i2c::{ErrorKind, NoAcknowledgeSource};
 use esp_idf_sys::*;
 
 use crate::{delay::*, gpio::*, units::*};
@@ -219,6 +221,27 @@ where
 
         Ok((self.i2c, self.pins))
     }
+
+    fn cmd_begin(
+        &mut self,
+        command_link: &CommandLink,
+        timeout: TickType_t,
+    ) -> Result<(), EspError> {
+        esp!(unsafe { i2c_master_cmd_begin(I2C::port(), command_link.0, timeout) })
+    }
+
+    fn submit(&mut self, command_link: &CommandLink) -> Result<(), I2cError> {
+        if let Err(err) = self.cmd_begin(command_link, self.timeout) {
+            let err = if err.code() == ESP_FAIL {
+                I2cError::new(ErrorKind::NoAcknowledge(NoAcknowledgeSource::Unknown), err)
+            } else {
+                I2cError::other(err)
+            };
+            Err(err)
+        } else {
+            Ok(())
+        }
+    }
 }
 
 impl<I2C, SDA, SCL> embedded_hal_0_2::blocking::i2c::Read for Master<I2C, SDA, SCL>
@@ -277,103 +300,67 @@ where
     SCL: OutputPin,
 {
     fn read(&mut self, addr: u8, buffer: &mut [u8]) -> Result<(), Self::Error> {
-        let command_link = CommandLink::new().map_err(I2cError::other)?;
+        let mut command_link = CommandLink::new().map_err(I2cError::other)?;
 
-        unsafe {
-            esp!(i2c_master_start(command_link.0)).map_err(I2cError::other)?;
-            esp!(i2c_master_write_byte(
-                command_link.0,
-                (addr << 1) | (i2c_rw_t_I2C_MASTER_READ as u8),
-                true
-            ))
+        command_link.master_start().map_err(I2cError::other)?;
+        command_link
+            .master_write_byte((addr << 1) | (i2c_rw_t_I2C_MASTER_READ as u8), true)
             .map_err(I2cError::other)?;
-            esp!(i2c_master_read(
-                command_link.0,
-                buffer.as_ptr() as *const u8 as *mut u8,
-                buffer.len() as u32,
-                i2c_ack_type_t_I2C_MASTER_LAST_NACK
-            ))
-            .map_err(I2cError::other)?;
-            esp!(i2c_master_stop(command_link.0)).map_err(I2cError::other)?;
-
-            esp_result!(
-                i2c_master_cmd_begin(I2C::port(), command_link.0, self.timeout),
-                ()
-            )
-            .map_err(I2cError::other)
+        if !buffer.is_empty() {
+            command_link
+                .master_read(buffer, AckType::LastNack)
+                .map_err(I2cError::other)?;
         }
+        command_link.master_stop().map_err(I2cError::other)?;
+
+        self.submit(&command_link)
     }
 
     fn write(&mut self, addr: u8, bytes: &[u8]) -> Result<(), Self::Error> {
-        unsafe {
-            let command_link = CommandLink::new().map_err(I2cError::other)?;
+        let mut command_link = CommandLink::new().map_err(I2cError::other)?;
 
-            esp!(i2c_master_start(command_link.0)).map_err(I2cError::other)?;
-            esp!(i2c_master_write_byte(
-                command_link.0,
-                (addr << 1) | (i2c_rw_t_I2C_MASTER_WRITE as u8),
-                true
-            ))
-            .map_err(I2cError::other)?;
-            esp!(i2c_master_write(
-                command_link.0,
-                bytes.as_ptr() as *const u8 as *mut u8,
-                bytes.len() as u32,
-                true
-            ))
-            .map_err(I2cError::other)?;
-            esp!(i2c_master_stop(command_link.0)).map_err(I2cError::other)?;
+        command_link.master_start().map_err(I2cError::other)?;
 
-            esp_result!(
-                i2c_master_cmd_begin(I2C::port(), command_link.0, self.timeout),
-                ()
-            )
-            .map_err(I2cError::other)
+        command_link
+            .master_write_byte((addr << 1) | (i2c_rw_t_I2C_MASTER_WRITE as u8), true)
+            .map_err(I2cError::other)?;
+
+        if !bytes.is_empty() {
+            command_link
+                .master_write(bytes, true)
+                .map_err(I2cError::other)?;
         }
+        command_link.master_stop().map_err(I2cError::other)?;
+
+        self.submit(&command_link)
     }
 
     fn write_read(&mut self, addr: u8, bytes: &[u8], buffer: &mut [u8]) -> Result<(), Self::Error> {
-        let command_link = CommandLink::new().map_err(I2cError::other)?;
+        let mut command_link = CommandLink::new().map_err(I2cError::other)?;
 
-        unsafe {
-            esp!(i2c_master_start(command_link.0)).map_err(I2cError::other)?;
-            esp!(i2c_master_write_byte(
-                command_link.0,
-                (addr << 1) | (i2c_rw_t_I2C_MASTER_WRITE as u8),
-                true
-            ))
+        command_link.master_start().map_err(I2cError::other)?;
+        command_link
+            .master_write_byte((addr << 1) | (i2c_rw_t_I2C_MASTER_WRITE as u8), true)
             .map_err(I2cError::other)?;
-            esp!(i2c_master_write(
-                command_link.0,
-                bytes.as_ptr() as *const u8 as *mut u8,
-                bytes.len() as u32,
-                true
-            ))
-            .map_err(I2cError::other)?;
-
-            esp!(i2c_master_start(command_link.0)).map_err(I2cError::other)?;
-            esp!(i2c_master_write_byte(
-                command_link.0,
-                (addr << 1) | (i2c_rw_t_I2C_MASTER_READ as u8),
-                true
-            ))
-            .map_err(I2cError::other)?;
-            esp!(i2c_master_read(
-                command_link.0,
-                buffer.as_ptr() as *const u8 as *mut u8,
-                buffer.len() as u32,
-                i2c_ack_type_t_I2C_MASTER_LAST_NACK
-            ))
-            .map_err(I2cError::other)?;
-
-            esp!(i2c_master_stop(command_link.0)).map_err(I2cError::other)?;
-
-            esp_result!(
-                i2c_master_cmd_begin(I2C::port(), command_link.0, self.timeout),
-                ()
-            )
-            .map_err(I2cError::other)
+        if !bytes.is_empty() {
+            command_link
+                .master_write(bytes, true)
+                .map_err(I2cError::other)?;
         }
+
+        command_link.master_start().map_err(I2cError::other)?;
+        command_link
+            .master_write_byte((addr << 1) | (i2c_rw_t_I2C_MASTER_READ as u8), true)
+            .map_err(I2cError::other)?;
+        if !buffer.is_empty() {
+            command_link
+                .master_read(buffer, AckType::LastNack)
+                .map_err(I2cError::other)?;
+        }
+
+        command_link.master_stop().map_err(I2cError::other)?;
+
+        self.submit(&command_link)
     }
 
     fn write_iter<B>(&mut self, _address: u8, _bytes: B) -> Result<(), Self::Error>
@@ -397,10 +384,61 @@ where
 
     fn transaction<'a>(
         &mut self,
-        _address: u8,
-        _operations: &mut [embedded_hal::i2c::blocking::Operation<'a>],
+        address: u8,
+        operations: &mut [embedded_hal::i2c::blocking::Operation<'a>],
     ) -> Result<(), Self::Error> {
-        todo!()
+        use embedded_hal::i2c::blocking::Operation;
+
+        let mut command_link = CommandLink::new().map_err(I2cError::other)?;
+
+        command_link.master_start().map_err(I2cError::other)?;
+
+        let last_op_index = operations.len() - 1;
+        let mut prev_was_read = None;
+
+        for (i, operation) in operations.iter_mut().enumerate() {
+            match operation {
+                Operation::Read(buf) => {
+                    if let Some(false) = prev_was_read {
+                        command_link.master_start().map_err(I2cError::other)?;
+                    }
+                    prev_was_read = Some(true);
+
+                    command_link
+                        .master_write_byte((address << 1) | (i2c_rw_t_I2C_MASTER_READ as u8), true)
+                        .map_err(I2cError::other)?;
+                    if !buf.is_empty() {
+                        let ack = if i == last_op_index {
+                            AckType::LastNack
+                        } else {
+                            AckType::Ack
+                        };
+                        command_link
+                            .master_read(*buf, ack)
+                            .map_err(I2cError::other)?;
+                    }
+                }
+                Operation::Write(buf) => {
+                    if let Some(true) = prev_was_read {
+                        command_link.master_start().map_err(I2cError::other)?;
+                    }
+                    prev_was_read = Some(false);
+
+                    command_link
+                        .master_write_byte((address << 1) | (i2c_rw_t_I2C_MASTER_WRITE as u8), true)
+                        .map_err(I2cError::other)?;
+                    if !buf.is_empty() {
+                        command_link
+                            .master_write(buf, true)
+                            .map_err(I2cError::other)?;
+                    }
+                }
+            }
+        }
+
+        command_link.master_stop().map_err(I2cError::other)?;
+
+        self.submit(&command_link)
     }
 
     fn transaction_iter<'a, O>(&mut self, _address: u8, _operations: O) -> Result<(), Self::Error>
@@ -519,9 +557,17 @@ where
     }
 }
 
-struct CommandLink(i2c_cmd_handle_t);
+#[repr(u32)]
+enum AckType {
+    Ack = i2c_ack_type_t_I2C_MASTER_ACK,
+    #[allow(dead_code)]
+    Nack = i2c_ack_type_t_I2C_MASTER_NACK,
+    LastNack = i2c_ack_type_t_I2C_MASTER_LAST_NACK,
+}
 
-impl CommandLink {
+struct CommandLink<'buffers>(i2c_cmd_handle_t, PhantomData<&'buffers u8>);
+
+impl<'buffers> CommandLink<'buffers> {
     fn new() -> Result<Self, EspError> {
         let handle = unsafe { i2c_cmd_link_create() };
 
@@ -529,11 +575,45 @@ impl CommandLink {
             return Err(EspError::from(ESP_ERR_NO_MEM as i32).unwrap());
         }
 
-        Ok(CommandLink(handle))
+        Ok(CommandLink(handle, PhantomData))
+    }
+
+    fn master_start(&mut self) -> Result<(), EspError> {
+        esp!(unsafe { i2c_master_start(self.0) })
+    }
+
+    fn master_stop(&mut self) -> Result<(), EspError> {
+        esp!(unsafe { i2c_master_stop(self.0) })
+    }
+
+    fn master_write_byte(&mut self, data: u8, ack_en: bool) -> Result<(), EspError> {
+        esp!(unsafe { i2c_master_write_byte(self.0, data, ack_en) })
+    }
+
+    fn master_write(&mut self, buf: &'buffers [u8], ack_en: bool) -> Result<(), EspError> {
+        esp!(unsafe {
+            i2c_master_write(
+                self.0,
+                buf.as_ptr() as *const u8 as *mut u8,
+                buf.len() as u32,
+                ack_en,
+            )
+        })
+    }
+
+    fn master_read(&mut self, buf: &'buffers mut [u8], ack: AckType) -> Result<(), EspError> {
+        esp!(unsafe {
+            i2c_master_read(
+                self.0,
+                buf.as_ptr() as *const u8 as *mut u8,
+                buf.len() as u32,
+                ack as u32,
+            )
+        })
     }
 }
 
-impl Drop for CommandLink {
+impl<'buffers> Drop for CommandLink<'buffers> {
     fn drop(&mut self) {
         unsafe {
             i2c_cmd_link_delete(self.0);
